
dyn.load("refineLikelihoodFastNu.so")

sigest_tnu <- function (x, frac = 0.5) {
  x <- as.matrix(x)
  m <- dim(x)[1]
  n <- floor(frac*m)
  index <- sample(1:m, n, replace = TRUE)
  index2 <- sample(1:m, n, replace = TRUE)
  temp <- x[index,, drop=FALSE] - x[index2,,drop=FALSE]
  dist <- rowSums(temp^2)
  srange <- 1/quantile(dist[dist!=0],probs=c(0.9,0.5,0.1))
  return(mean(srange[c(1,3)]))
}

source("gFITC.R")

source("copulaDensities.R")

likelihoodFactor_tnu <- function(a, x, y, b) {

	nu <- 1 + 1e6 * pnorm(a)
	tau <- 0.99 * (2 * pnorm(b) - 1)
	dTC(x, y, tau, nu)
}
	
epIDCinternal_tnu <- function(X, Z, Zbar, sigma, sigma0, l, m0, tauValues, start = NULL) {

	# We initialize the structure with the problem information

	gFITCinfo <- initGFITCinfo(Z, Zbar, sigma, sigma0, l, m0)

	# We initialize the approximate factors

	f1Hat <- list(eta1 = rep(1e-100, gFITCinfo$n), eta2 = rep(1e-100, gFITCinfo$n), logZ = rep(0, gFITCinfo$n))
	f2Hat <- list(eta1 = rep(1e-100, gFITCinfo$n), eta2 = rep(1e-100, gFITCinfo$n))

	# We initialize the posterior approximation

	a <- list(eta1 = rep(0, gFITCinfo$n), eta2 = rep(0, gFITCinfo$n))

	# We refine the second approximate factor

	a$eta2 <- f2Hat$eta2 <- gFITCinfo$diagKnn^-1
	a$eta1 <- f2Hat$eta1 <- m0 / gFITCinfo$diagKnn

	# We check for an initial solution

	if (!is.null(start)) {
		a <- start$a
		f1Hat <- start$f1Hat
		f2Hat <- start$f2Hat
	}

	# Main loop of EP

	i <- 1
	damping <- 1
	convergence <- FALSE
	while (!convergence && i < 50) {

		aOld <- a

		# We refine the first approximate factor
		
		f1HatOld <- f1Hat
		repeat {

			f1Hat <- f1HatOld

			# We create the auxiliary list m

			m <- list(X = X[ , 1 ], Y = X[ , 2 ],
				ma = tauValues$m, va = tauValues$v,
				mb = f2Hat$eta1 / f2Hat$eta2, vb = f2Hat$eta2^-1, 
				f1HatEta1 = f1Hat$eta1, f1HatEta2 = f1Hat$eta2,
				damping = damping, n = gFITCinfo$n, logZ = f1Hat$logZ)

			m <- .Call("refineLikelihoodFastNu", m, dup = F)

			f1Hat$eta1 <- m$f1HatEta1
			f1Hat$eta2 <- m$f1HatEta2
			f1Hat$logZ <- m$logZ

			if (checkPDposterior(gFITCinfo, f1Hat$eta2))
				break
			else
				damping <- damping * 0.5
		}

		# We refine the second approximate factor.

		ret <- computeTitledDistribution(gFITCinfo, f1Hat$eta2, f1Hat$eta1)
		eta1HatNew <- ret$mNew / ret$vNew - f1Hat$eta1
		eta2HatNew <- ret$vNew^-1 - f1Hat$eta2

		index <- which(is.infinite(eta2HatNew))
		if (length(index) > 0) {
			eta2HatNew[ index ] <- f2Hat$eta2[ index ]
			eta1HatNew[ index ] <- f2Hat$eta1[ index ]
		}
		
		index <- which(is.infinite(eta1HatNew))
		if (length(index) > 0) {
			eta2HatNew[ index ] <- f2Hat$eta2[ index ]
			eta1HatNew[ index ] <- f2Hat$eta1[ index ]
		}

		index <- which(damping * eta2HatNew + (1 - damping) * f2Hat$eta2 + f1Hat$eta2 < 0)
		if (length(index) > 0) {
			eta2HatNew[ index ] <- f2Hat$eta2[ index ]
			eta1HatNew[ index ] <- f2Hat$eta1[ index ]
		}
	
		eta1HatNew[ eta2HatNew <= 0 ] <- f2Hat$eta1[ eta2HatNew <= 0 ]
		eta2HatNew[ eta2HatNew <= 0 ] <- f2Hat$eta2[ eta2HatNew <= 0 ]

		f2Hat$eta1 <- damping * eta1HatNew + (1 - damping) * f2Hat$eta1
		f2Hat$eta2 <- damping * eta2HatNew + (1 - damping) * f2Hat$eta2

		# We update the posterior approximation

		a$eta1 <- f1Hat$eta1 + f2Hat$eta1
		a$eta2 <- f1Hat$eta2 + f2Hat$eta2

		# We check for convergence

		change <- max(abs(aOld$eta1 / aOld$eta2 - a$eta1 / a$eta2))
		change <- max(change, abs(aOld$eta2^-1 - a$eta2^-1))
		if (change < 1e-3)
			convergence <- T

#		cat("Iteration",  i, change, "\n")

		# Annealed damping scheme

		damping <- damping * 0.95

		i <- i + 1
	}

	# We compute the evidence and its gradient

	logZ <- computeEvidence_tnu(a, f1Hat, f2Hat, gFITCinfo, X)
	gradientLogZ <- computeDerivativesEvidence(gFITCinfo, f1Hat$eta2,
		f1Hat$eta1)

	# We are done!

	list(a = a, f1Hat = f1Hat, f2Hat = f2Hat, logZ = logZ,
		gradientLogZ = gradientLogZ, gFITCinfo = gFITCinfo,
		sigma = sigma, sigma0 = sigma0, l = l, m0 = m0, Zbar = Zbar)
}

computeEvidence_tnu <- function(a, f1Hat, f2Hat, gFITCinfo, X) {

	logZ <- sum(f1Hat$logZ)
	logZret <- logZ + sum(0.5 * log(2 * pi) -
		0.5 * log(f2Hat$eta2) + 0.5 * log(a$eta2) -
		0.5 * a$eta1^2 / a$eta2 + 0.5 * f1Hat$eta1^2 /
		f1Hat$eta2 + 0.5 * f2Hat$eta1^2 / f2Hat$eta2)

	logZret <- logZret + getFITCevidence(gFITCinfo, f1Hat$eta2, f1Hat$eta1)

	logZret
}

##
# Function that predicts for a new data point
#

predictIDC_tnu <- function(ret, zNew, n.samples) {

	retInternal <- predictFITC(ret$gFITCinfo, ret$f1Hat$eta2, ret$f1Hat$eta1, zNew)
	mZnew <- retInternal$m
	vZnew <- retInternal$v

	t(apply(cbind(mZnew, vZnew), 1, function(x) 1 + 1e6 * pnorm(rnorm(n.samples, x[ 1 ], sqrt(x[ 2 ])))))
}

##
# Function that computes an integral using the Newton-Cotes formula.
#

integrateFast_tnu <- function(f, minx, maxx) {

        x <- seq(minx, maxx, length.out = 41)
        h <- x[ 2 ] - x[ 1 ]

        # We evaluate the function to be integrated

        fvalues1 <- f(x)
        fvalues2 <- t.default(matrix(fvalues1[ 1 : (length(fvalues1) - 1) ], ncol = length(x) / 4, nrow = 4))
        fvalues3 <- cbind(fvalues2, c(fvalues2[ 2 : nrow(fvalues2) , 1 ], fvalues1[ length(fvalues1) ]))
        aux <- fvalues3 %*% c(7, 32, 12, 32, 7) * 2 * h / 45
        x2 <- seq(x[ 5 ], x[ length(x) ], 4 * h)
        sum(aux)
}

epIDCexternal_tnu <- function(X, Z, meanNu, tauValues, m, ret = NULL) {

	if (meanNu > 1e6)
		meanNu <- 1e6 - 1

	m0 <- qnorm((meanNu - 1) / 1e6)

	# We initialize the hyper-parameters if no initial solution given

	if (is.null(ret)) {

		sigma <- 0
		sigma0 <- -10
		l <- log(sqrt(apply(Z,2,sigest_tnu)))
		Zbar <- matrix(Z[ sample(1 : nrow(Z), m), ], m, ncol(Z))
		eps <- 0.1

	} else {

		sigma <- ret$sigma
		sigma0 <- ret$sigma0
		l <- ret$l
		Zbar <- ret$Zbar
		eps <- ret$eps

	}

	# We initialize the gradient optimization process

	resulttry <- try(retNew <- epIDCinternal_tnu(X, Z, Zbar, sigma, sigma0, l, m0, tauValues, ret))
	if (class(resulttry) == "try-error")
		return(ret)

#	if (!is.null(ret) && (retNew$logZ - ret$logZ) / abs(ret$logZ) > 20)
#		return(ret)

	ret <- retNew

	bestSoFar <- ret
	bestSoFar$eps <- eps

	cat(0, "New evidence:", ret$logZ, "\n")

	convergence <- F
	iteration <- 1
	while (!convergence && iteration < 50) {

		sigmaNew <- sigma + eps * ret$gradientLogZ$dLogZdSigma
		sigma0New <- sigma0 + eps * ret$gradientLogZ$dLogZdSigma0
		lNew <- l + eps * ret$gradientLogZ$dLogZdl
		ZbarNew <- Zbar + eps * ret$gradientLogZ$dLogZdXbar

		resulttry <- try(gFITCinfo <- initGFITCinfo(Z, ZbarNew, sigmaNew, sigma0New, lNew, m0))
		if (class(resulttry) == "try-error")
			return(bestSoFar)

		counter <- 1
		while (is.null(gFITCinfo) || !checkPDposterior(gFITCinfo, ret$f1Hat$eta2)) {
			eps <- eps * 0.5
			sigmaNew <- sigma + eps * ret$gradientLogZ$dLogZdSigma
			sigma0New <- sigma0 + eps * ret$gradientLogZ$dLogZdSigma0
			lNew <- l + eps * ret$gradientLogZ$dLogZdl
			ZbarNew <- Zbar + eps * ret$gradientLogZ$dLogZdXbar

			resulttry <- try(gFITCinfo <- initGFITCinfo(Z, ZbarNew, sigmaNew, sigma0New, lNew, m0))
			if (class(resulttry) == "try-error")
				return(bestSoFar)

			if (counter > 1e2) {
				return(bestSoFar)
			}

			counter <- counter + 1
		}
		sigma <- sigmaNew
		sigma0 <- sigma0New
		l <- lNew
		Zbar <- ZbarNew

		resulttry <- try(retNew <- epIDCinternal_tnu(X, Z, Zbar, sigma, sigma0, l, m0, tauValues, ret))
		if (class(resulttry) == "try-error")
			return(bestSoFar)

		if (abs(retNew$logZ - ret$logZ) / abs(ret$logZ) < 0.5 * 1e-4)
			convergence <- T

		if (retNew$logZ < ret$logZ)
			eps <- eps * 0.5
		else
			eps <- eps * 1.1

		if (retNew$logZ > bestSoFar$logZ) {
			bestSoFar <- retNew
			bestSoFar$eps <- eps
		}

		cat(iteration, "New evidence:", retNew$logZ, "eps:", eps, "\n")

		ret <- retNew

		iteration <- iteration + 1
	}

	bestSoFar
}
