
# We load the Bayesian quadrature module

source("epIDCttau.R")

source("epIDCtnu.R")

fitUTC <- function(X) {

	ll <- function(parameters) -sum(logdTC(X[ , 1 ], X[ , 2 ], rep(parameters[ 1 ], nrow(X)), rep(parameters[ 2 ], nrow(X))))

	ret <- optim(c(0, 10), ll, method = "L-BFGS-B", lower = c(-0.99, 1), upper = c(0.99, 1e6), control = list(trace = T))$par

	list(tau = ret[ 1 ], nu = ret[ 2 ], value = -ll(c(ret)))
}

fitCTC <- function(X, Z, m = round(nrow(X) / 10)) {
	
	# We fit an unconditional t coupla to the data

	retOptimizer <- fitUTC(X)

	retNu <- retOptimizer$nu
	retTau <- retOptimizer$tau
	logZ <- retOptimizer$value

	list(retNu = retNu, retTau = retTau, logZsequence = logZ)
}

sampleCTC <- function(n, ret, Ztest) {

	sample <- matrix(0, n, 2)
	tau <- as.double(predictIDC_ttau(ret$retTau, matrix(Ztest, 1, length(Ztest), n)))
	nu <- as.double(predictIDC_tnu(ret$retNu, matrix(Ztest, 1, length(Ztest), n)))

	for (i in 1 : n)
		sample[ i , ] <- rCopula(1, tCopula(sin((tau[ i ] * pi) / 2), 2, df = nu[ i ]))

	sample
}

evaluateLLCTC <- function(ret, Xtest, Ytest, Ztest, n.samples = 100000) {

	# We generate predictions for each datapoint

	tau <- matrix(ret$retTau, nrow(Ztest), 1)
	nu <- matrix(ret$retNu, nrow(Ztest), 1)

	Xtest <- matrix(Xtest, nrow(tau), ncol(tau))
	Ytest <- matrix(Ytest, nrow(tau), ncol(tau))

	ll <- log(dTC(Xtest, Ytest, tau, nu) %*% rep(1 / ncol(tau), ncol(tau)))

	# We obtain the mean and the lower quantiles for tau and nu

	medianTau <- apply(tau, 1, median)
	quantileInfTau <- apply(tau, 1, function(x) quantile(x, 0.1))
	quantileSupTau <- apply(tau, 1, function(x) quantile(x, 0.9))

	medianNu <- apply(nu, 1, median)
	quantileInfNu <- apply(nu, 1, function(x) quantile(x, 0.1))
	quantileSupNu <- apply(nu, 1, function(x) quantile(x, 0.9))

	# We evaluate the test log likelihood on each data point

	list(ll = ll, medianTau = medianTau, quantileInfTau = quantileInfTau, quantileSupTau = quantileSupTau,
		medianNu = medianNu, quantileSupNu = quantileSupNu, quantileInfNu = quantileInfNu)
}
